#!/bin/python3
#%%
import pandas as pd
import numpy as np
from tools.mdl_kit_v6 import  check_dataframe, dict_to_json, get_int_from_str, auto_reader, t_print, count_down, stat_dataframe
from tools.y_funcs import merge_to_y_idx
from p9_funcs import get_y_p9


def get_qcut_bins(series: pd.Series, q = 10, mode = None):
    series = series[~series.isna()]
    flag = True
    if mode == 'auto':
        uniques = sorted(series.unique().tolist())
        if len(uniques) <= 10:
            bins = uniques
            t_print(f'uniques = {uniques}')
            flag = False
        else:
            cnt_sr = series.value_counts()
            cnt_sr = cnt_sr / cnt_sr.sum()
            print(cnt_sr)
            bins1 = cnt_sr[cnt_sr >= .05].index.to_list()
            bins2 = cnt_sr[cnt_sr < .05].index.to_list()
            new_series = series[series.isin(bins2)]
            if len(new_series) > 0:
                new_bins = pd.qcut(new_series, q=q, duplicates='drop', retbins=True)[1].tolist()
                bins = sorted(set(bins1) | set(new_bins))
            else:
                bins = sorted(bins1)
            flag = False
        # else:
        #     flag = True
    if flag:
        bins = pd.qcut(series, q=q, duplicates='drop', retbins=True)[1]
        bins = bins.tolist()
    return bins


def cal_bins_and_cnts(series: pd.Series, bins = 10, percent=False, ret = 'series'):
    if type(bins) == int and bins > 0:
        bins = get_qcut_bins(series, q = bins)
    bins = [-float('inf')] + bins + [float('inf')]
    cut_sr = pd.cut(series, bins = bins)
    cnt_sr = cut_sr.value_counts().sort_index()
    cnt_sr['Not_a_Number'] = series.isna().sum()
    if percent :
        cnt_sr = cnt_sr / cnt_sr.sum()
    r = cnt_sr
    if ret == 'dict':
        cnt_sr.index = cnt_sr.index.to_series().apply(lambda s: s.__str__())
        cnt_dict = cnt_sr.to_dict()
        r = cnt_dict
    return r


def cal_psi(a_cnt_sr, b_cnt_sr):
    a_cnt_sr = a_cnt_sr.copy()
    b_cnt_sr = b_cnt_sr.copy()
    a_cnt_sr /= a_cnt_sr.sum()
    b_cnt_sr /= b_cnt_sr.sum()
    for sr in [a_cnt_sr, b_cnt_sr]:
        for k, v in sr.items():
            if v == 0:
                sr[k] = 1e-6
    cnt_df = pd.DataFrame({'a': a_cnt_sr, 'b': b_cnt_sr})
    cnt_df['psi'] = (a_cnt_sr - b_cnt_sr) * np.log(a_cnt_sr / b_cnt_sr)
    psi = cnt_df['psi'].sum()
    print(cnt_df)
    print('psi =', psi)
    return cnt_df, psi


def get_src_features(batch: str, src: str):
    if src in [2,3,4,5,7]:
        src = f'S0{src}'
    # stat_path = f'./stat/step1/{batch}-{src}-step1-stat.pickle'
    # stat_df = pd.read_pickle(stat_path)
    # features = stat_df['col_name'][stat_df['to_mdl']]
    src_dict = {
        # 'S01': '',
        'S02': 'YS',
        'S03': 'JD',
        'S04': 'YK',
        'S05': 'TD',
        # 'S06': '',
        'S07': 'S07',
    }
    if src in src_dict:
        print(f'source: {src} -> {src_dict[src]}')
        src = src_dict[src]
    if src == 'YS':
        path = f'./data/s237_data/{batch}-S02-t8-step2-pickle'
    elif src == 'JD':
        path = f'./data/s237_data/{batch}-S03-a-step2-pickle'
    elif src == 'YK':
        path = f'./data/S04/{batch}-S04-origin-pickle'
    elif src == 'TD':
        path = f'./data/S05/{batch}-S05-fs_and_tn-pickle'
    elif src == 'S07':
        path = f'./data/s237_data/{batch}-S07-origin-pickle'
    t_print('read', path)
    df = pd.read_pickle(path)
    n1 = df.shape[1]
    df = df.select_dtypes(include = 'number')
    n2 = df.shape[1]
    t_print(f'cols: {n1} -> {n2}')
    t_print(df)
    return df


# def get_y_p9(batch: str) -> pd.DataFrame:
#     if batch in ['m22_api', 'p9_api', 'p9_b']: # 33w
#         path = './merchants/m22/data/hash/测试数据20240322-API.xlsx'
#     elif batch in ['m22_no', 'p9_no', 'p9_a']: # 10w
#         path = './merchants/m22/data/hash/测试数据样本-no.xlsx'
#     a_df = auto_reader(path, make_cache=True)
#     a_df.index = a_df['no'].apply(get_int_from_str)
#     a_df.index.name = 'index'
#     return a_df


def get_monthly_col_distri(col_sr, month_idx, col_bins):
    sr = col_sr.loc[month_idx]
    monthly_col_distri = cal_bins_and_cnts(sr, bins=col_bins)
    return monthly_col_distri


def get_col_distri_by_month(col_sr, month_index_dict, col_bins):
    month_col_dict = {}
    for month, month_idx in month_index_dict.items():
        monthly_col_distri = get_monthly_col_distri(col_sr, month_idx=month_idx, col_bins=col_bins)
        month_col_dict[month] = monthly_col_distri
    month_col_df = pd.DataFrame(month_col_dict)
    return month_col_df


def get_psi_by_month(col_distri, monthly_col_distri_df):
    monthly_col_psi_dict = {}
    for month in monthly_col_distri_df:
        month_distri = monthly_col_distri_df[month]
        monthly_psi = cal_psi(col_distri, month_distri)[1]
        monthly_col_psi_dict[month] = monthly_psi
    return monthly_col_psi_dict


def get_df_info_value(data_df: pd.DataFrame, y_sr: pd.Series):
    if isinstance(y_sr, pd.DataFrame):
        y_df = y_sr
        y_sr = y_df.loc[y_df['mdl_type'] == 'train', 'y']
    data_df = pd.merge(
        pd.DataFrame(index=y_sr.index), 
        data_df, 
        left_index=True, 
        right_index=True, 
        how='left' )
    rows_dict = {}
    k = 0
    for col in data_df:
        k += 1
        t_print('\n\n#', f'{k}/{data_df.shape[1]}', col)
        col_dict = {}
        col_sr = data_df[col]
        pos_col_sr = data_df.loc[y_sr==1, col]
        neg_col_sr = data_df.loc[y_sr==0, col]
        col_bins = get_qcut_bins(col_sr, mode = 'auto')
        pos_col_distri = cal_bins_and_cnts(pos_col_sr, bins=col_bins, )
        neg_col_distri = cal_bins_and_cnts(neg_col_sr, bins=col_bins, )
        psi = cal_psi(pos_col_distri, neg_col_distri)[1]
        col_dict['pos_dist'] = dict_to_json(pos_col_distri.to_dict())
        col_dict['neg_dist'] = dict_to_json(neg_col_distri.to_dict())
        col_dict['info_value'] = psi
    monthly_psi_df = pd.DataFrame(rows_dict)
    monthly_psi_df = monthly_psi_df.T
    t_print('end of get_df_info_value()')
    return monthly_psi_df




