#!/bin/python3
#%%
import pandas as pd
from tools.mdl_kit_v6 import auto_reader, get_mdl_type_sr, stat_ya_df

#%%



def get_hash_df(batch: str):
    if batch == 'p9_a':
        path = './data/y/final_md52_output_y2.csv'
        date_col = 'crd_dt'
        y_col = 'y2'
        test_date = '2023-12-01'
        stratify = [y_col, 'month']
    elif batch == 'p9_b':
        path = './data/y/联合建模测试样本2024-03-06.xlsx'
        date_col = 'apply_dt'
        y_col = 'mob3_30_flag'
        test_date = '2023-11-01'
        stratify = [y_col, 'month']
    print('date_col =', date_col)
    print('y_col = ', y_col)
    df = auto_reader(path)
    df['month'] = df[date_col].dt.strftime('%Y-%m')
    df['mdl_type'] = get_mdl_type_sr(
        ya_df=df, 
        y_col=y_col, 
        test_date=test_date,
        date_col=date_col, 
        stratify=stratify)
    stat_ya_df(df, y_col=y_col, date_col=date_col)
    return df

api_tag_dict = {
    'jd': 'API-京东', # 9.7
    'zj': 'API-字节', # 7.5 zj
    'xx': 'API-星选', # 2.3 zj
    'el': 'API-饿了么', # 2.1
    'zx': 'API-智信', # 1.3 360
}


def get_y_p9(batch: str, mode = 1):
    if batch in ['a', 'b']:
        batch = f'p9_{batch}'
    if batch not in ['p9_a'] and not batch.startswith('p9_b'):
        raise ValueError('batch not in p9_a, p9_b')
    if batch == 'p9_a':
        date_col = 'crd_dt'
        y_col = 'y2'
    elif batch.startswith('p9_b'):
        date_col = 'apply_dt'
        y_col = 'mob3_30_flag'
    batch_file = batch[:4]
    path = f'./data/y/{batch_file}-y-origin-pickle'
    df = pd.read_pickle(path)
    if batch.startswith('p9_b_'):
        api_tag = batch[-2:]
        if api_tag not in api_tag_dict:
            raise ValueError(f'api_not in {api_tag_dict}')
        api_tag = api_tag_dict[api_tag]
        idx = df['product2'] == api_tag
        df = df.loc[idx, :]
    if mode in [1, 2]:
        rename = {
            date_col: 'date',
            y_col: 'y'
        }
        df.rename(rename, axis = 1, inplace = True)
    if mode == 1:
        y_idx = ~df['y'].isna()
        ret = df.loc[y_idx, 'y']
    elif mode == 2:
        y_idx = ~df['y'].isna()
        ret = df.loc[y_idx, :]
    else:
        ret = df
    return ret


def get_src_features(batch: str, src: str, features = []):
    path = f'./data/com_data/{batch}-{src}-to_mdl_v1.pickle'
    df = pd.read_pickle(path)
    return df

