#!/bin/python3
#%%
import os, sys
import numpy as np
import pandas as pd
from tools.mdl_kit_v6 import auto_reader, t_print, get_auc_ks, json_dump

#%%

batch = sys.argv[1]

#%%
if batch == 'p9_a':
    y_col = 'y2'
    date_col = 'crd_dt'
    drop_cols = ['y2']
elif batch.startswith('p9_b'):
    y_col = 'mob3_30_flag'
    date_col = 'apply_dt'
    drop_cols = ['fpd7_flag', 'fpd30_flag', 'mob2_30_flag', 'mob3_30_flag']

#%%

to_long = {
    'p9_a':    'scene_y2',
    'p9_b':    'scene_m3_30-API_all',
    'p9_b_jd': 'scene_m3_30-API_jingdong',
    'p9_b_zj': 'scene_m3_30-API_zijie',
    'p9_b_xx': 'scene_m3_30-API_xingxuan',
    'p9_b_el': 'scene_m3_30-API_eleme',
    'p9_b_zx': 'scene_m3_30-API_zhixin',
}

def get_renamed_scores(batch: str):
    path = f'./output_files/scores/{batch}-model_scores.csv'
    df = auto_reader(path, index_col=0)
    t_print(list(df))
    score_cols = df.columns.str.startswith('p9_')
    score_cols = df.columns[score_cols]
    rename = {a: f'{to_long[batch]}-score_{k+1}' for k, a in enumerate(score_cols)}
    t_print('rename =', rename)
    df.rename(rename, axis=1, inplace=True)
    return df, rename

#%%

def get_monthly_sample_stat(df: pd.DataFrame):

    df = df.copy()
    df = df.loc[~df['mdl_type'].isna(), :]

    good_s = df.loc[df[y_col] == 0, :].groupby(by='month').size().sort_index()
    good_s.name = '好客户数'

    bad_s = df.loc[df[y_col] == 1, :].groupby(by='month').size().sort_index()
    bad_s.name = '坏客户数'

    all_s = df.loc[df[y_col].isin([0, 1]), :].groupby(by='month').size().sort_index()
    all_s.name = '样本量'

    bad_rat_s = bad_s / all_s
    bad_rat_s.name = '坏账率'

    sample_rat_s = all_s / all_s.sum()
    sample_rat_s.name = '样本占比'

    monthly_stat_df = pd.concat([good_s, bad_s, all_s, bad_rat_s, sample_rat_s], axis=1)
    monthly_stat_df = pd.concat([
        monthly_stat_df,
        pd.DataFrame({
            '好客户数': [good_s.sum()],
            '坏客户数': [bad_s.sum()],
            '样本量': [all_s.sum()],
            '坏账率': [bad_s.sum() / all_s.sum()],
            '样本占比': [1]
        }, index=['total'])
    ])

    monthly_stat_df.insert(0, '月份', monthly_stat_df.index)
    monthly_stat_df.index = pd.Index(range(1, len(monthly_stat_df)+1), name='index')
    print('# monthly_sample_stat')
    print(monthly_stat_df)
    print('\n\n')

    return monthly_stat_df


#%%

def get_sample_stat(df: pd.DataFrame):

    df = df.copy()
    mdl_df = df.loc[~df['mdl_type'].isna(), :]

    min_date_s = mdl_df.groupby(by='mdl_type')[date_col].min()
    max_date_s = mdl_df.groupby(by='mdl_type')[date_col].max()
    date_df = pd.concat([min_date_s, max_date_s], axis=1)
    date_s = date_df.apply(lambda s: '~'.join([
        d.strftime('%Y-%m-%d') for d in s
        ]), axis=1)
    date_s.name = '时间范围'

    good_s = df.loc[df[y_col] == 0, :].groupby(by='mdl_type').size()
    good_s.name = '好客户数'

    bad_s = df.loc[df[y_col] == 1, :].groupby(by='mdl_type').size()
    bad_s.name = '坏客户数'

    all_s = df.loc[df[y_col].isin([0, 1]), :].groupby(by='mdl_type').size()
    all_s.name = '样本量'

    bad_rat_s = bad_s / all_s
    bad_rat_s.name = '坏账率'

    sample_rat_s = all_s / all_s.sum()
    sample_rat_s.name = '样本占比'

    sampel_stat_df = pd.concat([
        date_s, good_s, bad_s, all_s, bad_rat_s, sample_rat_s
    ], axis=1)

    sampel_stat_df.rename({
        'train': 'train', 'validate': 'test', 'test': 'oot'
    }, inplace=True)

    sampel_stat_df = sampel_stat_df.loc[['train', 'test', 'oot'], :]
    sampel_stat_df.insert(0, '样本', sampel_stat_df.index, )
    sampel_stat_df.index = pd.Index([1,2,3], name='index')
    print(sampel_stat_df)

    return sampel_stat_df


#%%

def get_cut_points(s: pd.Series, bins = 10, endpoints = [0, 100]):
    points = []
    s = pd.Series(s).copy()
    s.fillna(s.mean(), inplace=True)
    s.sort_values(inplace=True)
    for k in range(1, bins):
        n = int(len(s) * k / bins)
        cut = s.iloc[n]
        points.append(cut)
    points = sorted(set(points + endpoints))
    t_print(points)
    return points


def cal_bins_and_cnts(series: pd.Series, bins = 10, percent=False, ret = 'series'):
    cut_sr = pd.cut(series, bins = bins)
    cnt_sr = cut_sr.value_counts().sort_index()
    cnt_sr['Not_a_Number'] = series.isna().sum()
    if percent :
        cnt_sr = cnt_sr / cnt_sr.sum()
    r = cnt_sr
    if ret == 'dict':
        cnt_sr.index = cnt_sr.index.to_series().apply(lambda s: s.__str__())
        cnt_dict = cnt_sr.to_dict()
        r = cnt_dict
    return r


def cal_psi(a_cnt_sr, b_cnt_sr):
    a_cnt_sr = a_cnt_sr.copy()
    b_cnt_sr = b_cnt_sr.copy()
    a_cnt_sr /= a_cnt_sr.sum()
    b_cnt_sr /= b_cnt_sr.sum()
    for sr in [a_cnt_sr, b_cnt_sr]:
        for k, v in sr.items():
            if v == 0:
                sr[k] = 1e-6
    cnt_df = pd.DataFrame({'a': a_cnt_sr, 'b': b_cnt_sr})
    cnt_df['psi'] = (a_cnt_sr - b_cnt_sr) * np.log(a_cnt_sr / b_cnt_sr)
    psi = cnt_df['psi'].sum()
    print(cnt_df)
    print('psi =', psi)
    return psi


def total_performance(df: pd.DataFrame, score_name: str):
    df = df.copy()
    mdl_df = df.loc[~df['mdl_type'].isna(), :]
    auc_ks_df = mdl_df.groupby(by='mdl_type').apply(
        lambda s: get_auc_ks(
            s[y_col], 
            s[score_name].fillna(s[score_name].mean()), 
            dtype='pd.Series'
            ))
    bins = get_cut_points(mdl_df[score_name])
    cnt_sr = cal_bins_and_cnts(mdl_df[score_name], bins=bins)
    auc_ks_df['psi'] = mdl_df.groupby(by='mdl_type').apply(
        lambda df: cal_psi(
            cnt_sr, 
            cal_bins_and_cnts(df[score_name], bins=bins))
            )
    all_auc_ks = get_auc_ks(
        mdl_df[y_col], mdl_df[score_name].fillna(mdl_df[score_name].mean()),
        dtype='pd.Series'
        )
    all_auc_ks['psi'] = .0
    all_auc_ks.name = 'all'
    all_auc_ks = all_auc_ks.to_frame().T
    print(auc_ks_df)
    print(all_auc_ks)
    auc_ks_df = pd.concat([auc_ks_df, all_auc_ks])
    auc_ks_df.rename({
        'train': 'train', 'validate': 'test', 'test': 'oot'
    }, inplace=True)

    auc_ks_df = auc_ks_df.loc[['train', 'test', 'oot', 'all'], :]
    auc_ks_df.insert(0, '数据集', auc_ks_df.index, )
    auc_ks_df.index = pd.Index([1,2,3,4], name='index')
    print(auc_ks_df)
    return auc_ks_df


#%%

def monthly_performance(df: pd.DataFrame, score_name: str):
    df = df.copy()
    mdl_df = df.loc[~df['mdl_type'].isna(), :]
    auc_ks_df = mdl_df.groupby(by='month').apply(
        lambda s: get_auc_ks(
            s[y_col], 
            s[score_name].fillna(s[score_name].mean()), 
            dtype='pd.Series'
            ))
    bins = get_cut_points(mdl_df[score_name])
    cnt_sr = cal_bins_and_cnts(mdl_df[score_name], bins=bins)
    auc_ks_df['psi'] = mdl_df.groupby(by='month').apply(
        lambda df: cal_psi(
            cnt_sr, 
            cal_bins_and_cnts(df[score_name], bins=bins))
            )
    auc_ks_df.insert(0, '月份', auc_ks_df.index,)
    auc_ks_df.sort_index(inplace = True)
    auc_ks_df.index = pd.Index(range(1, len(auc_ks_df)+1), name='index')
    print(auc_ks_df)
    return auc_ks_df


#%%

def get_lift_curve(y, y_proba, bins, prefix = ''):
    bins = sorted(bins, reverse=True)
    df = pd.DataFrame(
        index=range(1, len(bins)),
        columns=['interval', 'min_score', 'max_score', 'good_cnt', 'bad_cnt', 'cnt', 'bad_rate', 'lift']
        )
    y_mean = y.mean()
    for k, cut in enumerate(bins[1:]):
        k += 1
        idx = y_proba > cut
        cut_y = y[idx]
        cut_y_proba = y_proba[idx]
        df.loc[k, 'interval'] = f'({round(cut, 4)}, {max(bins)}]'
        df.loc[k, 'min_score'] = cut_y_proba.min()
        df.loc[k, 'max_score'] = cut_y_proba.max()
        df.loc[k, 'cnt'] = len(cut_y)
        df.loc[k, 'good_cnt'] = (cut_y == 0).sum()
        df.loc[k, 'bad_cnt'] = (cut_y == 1).sum()
        df.loc[k, 'bad_rate'] = df.loc[k, 'bad_cnt'] / df.loc[k, 'cnt']
        df.loc[k, 'lift'] = df.loc[k, 'bad_rate'] / y_mean
    rename = {
        'interval': '分数区间',
        'min_score': '区间最小值',
        'max_score': '区间最大值',
        'good_cnt': '好客户数',
        'bad_cnt': '坏客户数',
        'cnt': '样本量',
        'bad_rate': '坏账率',
        'lift': 'lift'
        }
    df.rename(rename, axis=1, inplace=True)
    print(df)
    return df


def get_split_lift(df: pd.DataFrame, score_name: str, split: str):
    t_print(f'\n\nsplit = {split}')
    y = df.loc[df['mdl_type'] == split, y_col]
    y_proba = df.loc[df['mdl_type'] == split, score_name]
    y_proba.fillna(y_proba.mean(), inplace=True)
    bins = get_cut_points(y_proba, bins=20, endpoints=[0, 100])
    split_df = get_lift_curve(y=y, y_proba=y_proba, bins=bins)
    return split_df


def get_tri_lift(df: pd.DataFrame, score_name: str):
    df = df.copy()
    train_df = get_split_lift(df, score_name, 'train')
    train_df.insert(0, '样本', 'train')
    validate_df = get_split_lift(df, score_name, 'validate')
    validate_df.insert(0, '样本', 'validate')
    test_df = get_split_lift(df, score_name, 'test')
    test_df.insert(0, '样本', 'test')
    stat_df = pd.concat([train_df, validate_df, test_df])
    replace = {
        'train': 'train',
        'validate': 'test',
        'test': 'oot',
    }
    stat_df['样本'] = stat_df['样本'].replace(replace)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')
    return stat_df


#%%

def get_split_distribution(df: pd.DataFrame, score_name: str, bins: list, split: str, prefix = ''):
    t_print(f'\n\nsplit = {split}')
    y_proba = df.loc[df['mdl_type'] == split, score_name]
    y_proba.fillna(y_proba.mean(), inplace=True)
    cnt_s = pd.cut(y_proba, bins=bins).value_counts().sort_index()
    distr_s = cnt_s / cnt_s.sum()
    stat_df = pd.DataFrame({'cnt': cnt_s, 'distri': distr_s})
    if prefix == 'split':
        prefix = split + '_'
    stat_df = stat_df.add_prefix(prefix=prefix)
    print(stat_df)
    return stat_df


def get_tri_distribution(df: pd.DataFrame, score_name: str):
    df = df.copy()
    bins = get_cut_points(
        df.loc[~ df['mdl_type'].isna(), score_name], 
        bins=10, 
        endpoints=[0, 100]
        )
    train_df = get_split_distribution(df, score_name=score_name, bins=bins, split='train', prefix = 'split')
    validate_df = get_split_distribution(df, score_name=score_name, bins=bins, split='validate', prefix = 'split')
    test_df = get_split_distribution(df, score_name=score_name, bins=bins, split='test', prefix = 'split')
    stat_df = pd.concat([train_df, validate_df, test_df], axis=1)
    stat_df.insert(0, 'interval', stat_df.index)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')
    stat_df = stat_df[['interval', 'train_distri', 'validate_distri', 'test_distri']]
    stat_df.rename({
        'interval': '区间',
        'train_distri': 'train',
        'validate_distri': 'test',
        'test_distri': 'oot',
    }, axis=1, inplace=True)
    print(stat_df)
    return stat_df


#%%

def save_dataframes_to_excel(xlsx_dict: dict, path: str, index = True):
    with pd.ExcelWriter(path) as writer:
        for sheet_name, df in xlsx_dict.items():
            df.to_excel(writer, sheet_name=sheet_name, index=index)
    print(f'xlsx file saved successfully: {path}')
    return None

#%%
df, rename = get_renamed_scores(batch)

#%%

data_split_replace = {
        'train': 'train',
        'validate': 'test',
        'test': 'oot',
    }

dir_path = f'./output_files/scores_and_reports/'
os.makedirs(dir_path, exist_ok=True)

csv_name = f'{to_long[batch]}-scores.csv'
csv_path = os.path.join(dir_path, csv_name)
t_print('saving to', csv_path)
df.rename({'mdl_type': 'data_split'}, axis=1).replace(data_split_replace).to_csv(csv_path)

#%%
csv100_name = f'{to_long[batch]}-scores-head100.csv'
csv100_path = os.path.join(dir_path, csv100_name)
df.drop(columns=drop_cols).head(100).rename({'mdl_type': 'data_split'}, axis=1).replace(data_split_replace).to_csv(csv100_path)

#%%
print(df)
os.makedirs('./output_files/rename/', exist_ok=True)
json_dump(rename, f'./output_files/rename/{batch}-com-LR-rename.json')

#%%

def save_report(df: pd.DataFrame, score_name: str):
    xlsx_dict = {}
    xlsx_dict['0.样本分析'] = get_monthly_sample_stat(df)
    xlsx_dict['1.1.样本概况'] = get_sample_stat(df)
    xlsx_dict['1.2.整体效果'] = total_performance(df, score_name)
    xlsx_dict['1.3.模型稳定性(分月ks,auc,psi)'] = monthly_performance(df, score_name)
    xlsx_dict['1.4.lift'] = get_tri_lift(df=df, score_name=score_name)
    xlsx_dict['1.5.评分分布'] = get_tri_distribution(df=df, score_name=score_name)
    dir_path = f'./output_files/scores_and_reports/{to_long[batch]}-reports/'
    os.makedirs(dir_path, exist_ok=True)
    xlsx_name = f'{score_name}.xlsx'
    path = os.path.join(dir_path, xlsx_name)
    save_dataframes_to_excel(xlsx_dict, path = path)
    return None


score_name_ls = list(rename.values())
# score_name = score_name_ls[0]

for score_name in score_name_ls:
    save_report(df, score_name)
    del score_name

#%%
print('end')
